#include <stdio.h>
#include <CL/opencl.h>

#define SOURCE_FILE_NAME "addvector.cl"
#define OPENCL_COMPILER_OPTIONS "-cl-mad-enable"
#define TEST_VECTOR_SIZE 1024
#define WORK_GROUP_SIZE 1
#define quit(msg) {printf(msg "\n");return(1);}

int main(int argc, char** argv)
{
	cl_int ocl_error;

	//Query platform count
	cl_uint num_platforms;
	if (clGetPlatformIDs(0, NULL, &num_platforms) != CL_SUCCESS) quit("Error getting OpenCL Platform Count");
	if (num_platforms == 0) quit("No OpenCL Platform found");
	printf("%d OpenCL Platforms found\n", num_platforms);
	
	//Query platforms
	cl_platform_id* platforms = new cl_platform_id[num_platforms];
	if (platforms == NULL) quit("Memory allocation error");
	if (clGetPlatformIDs(num_platforms, platforms, NULL) != CL_SUCCESS) quit("Error getting OpenCL Platforms");

	//Query and display platform information
	for (int i = 0;i < num_platforms;i++)
	{
		char platform_profile[64], platform_version[64], platform_name[64], platform_vendor[64];
		clGetPlatformInfo(platforms[i], CL_PLATFORM_PROFILE, 64, platform_profile, NULL);
		clGetPlatformInfo(platforms[i], CL_PLATFORM_VERSION, 64, platform_version, NULL);
		clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, 64, platform_name, NULL);
		clGetPlatformInfo(platforms[i], CL_PLATFORM_VENDOR, 64, platform_vendor, NULL);
		printf("Platform %d: (%s %s) %s %s\n", i, platform_profile, platform_version, platform_vendor, platform_name);
	}

	//Set first platform as default platform
	cl_platform_id platform = platforms[0];
	delete[] platforms;

	//Query device count for this platform
	cl_uint num_devices;
	clGetDeviceIDs(platform, CL_DEVICE_TYPE_ALL, 0, NULL, &num_devices);
	if (num_devices == 0) quit("No OpenCL device for this platform found");
	printf("%d OpenCL devices found for this platform\n", num_devices);

	//Query devices
	cl_device_id* devices = new cl_device_id[num_devices];
	if (devices == NULL) quit("Memory allocation error");
	if (clGetDeviceIDs(platform, CL_DEVICE_TYPE_ALL, num_devices, devices, NULL) != CL_SUCCESS) quit("Error getting OpenCL devices");

	//Query and display device information
	for (int i = 0;i < num_devices;i++)
	{
		char device_vendor[64], device_name[64];
		clGetDeviceInfo(devices[i], CL_DEVICE_NAME, 64, device_name, NULL);
		clGetDeviceInfo(devices[i], CL_DEVICE_VENDOR, 64, device_vendor, NULL);
		printf("Device %d: %s %s\n", i, device_vendor, device_name);
	}

	//Set first device as default device
	cl_device_id device = devices[0];
	delete[] devices;

	//Create OpenCL context
	cl_context context = clCreateContext(NULL, 1, &device, NULL, NULL, &ocl_error);
	if (ocl_error != CL_SUCCESS) quit("Error creating OpenCL context");

	//Create OpenCL command_queue
	cl_command_queue command_queue = clCreateCommandQueue(context, device, 0, &ocl_error);
	if (ocl_error != CL_SUCCESS) quit("Error creating OpenCL command queue");

	//Allocate host and device memory for three vectors
	int *host_vector[3];			//vector[0] and vector[1] are the source for the addition, result stored in vector[2]
	cl_mem device_vector[3];
	for (int i = 0;i < 3;i++)
	{
		if ((host_vector[i] = new int[TEST_VECTOR_SIZE]) == NULL) quit("Memory allocation error");
		device_vector[i] = clCreateBuffer(context, CL_MEM_READ_WRITE, TEST_VECTOR_SIZE * sizeof(int), NULL, &ocl_error);
		if (ocl_error != CL_SUCCESS) quit("Error allocating device memory");
	}

	//Initialize vectors
	for (int i = 0;i < TEST_VECTOR_SIZE;i++)
	{
		host_vector[0][i] = host_vector[1][i] = i;
		host_vector[2][i] = 0;
	}

	//Copy vectors to device
	for (int i = 0;i < 3;i++)
		if (clEnqueueWriteBuffer(command_queue, device_vector[i], CL_FALSE, 0, TEST_VECTOR_SIZE * sizeof(int), host_vector[i], 0, NULL, NULL) != CL_SUCCESS) quit("Error copying vectors to device");

	//Read Sourcecode from file
	unsigned int filesize;
	FILE* pFile;
	//Open file
	if ((pFile = fopen(SOURCE_FILE_NAME, "rb")) == NULL) quit("Error opening OpenCL source code file");
	//Optain File Size
	fseek (pFile , 0 , SEEK_END);
	filesize = ftell (pFile);
	rewind (pFile);
	//Read file
	char* sourceCode = new char[filesize + 1];
	if (fread(sourceCode, 1, filesize, pFile) != filesize) quit("Error reading source file");
	//Make sourceCode 0-terminated
	sourceCode[filesize] = 0;
	fclose(pFile);

	//Create OpenCL program object
	cl_program program = clCreateProgramWithSource(context, 1, (const char**) &sourceCode, NULL, &ocl_error);
	if (ocl_error != CL_SUCCESS) quit("Error creating program object");
	delete[] sourceCode;

	//Compile program
	ocl_error = clBuildProgram(program, 1, &device, OPENCL_COMPILER_OPTIONS, NULL, NULL);
	if (ocl_error != CL_SUCCESS) quit("Error compiling program");

	//Create kernel
	cl_kernel kernel = clCreateKernel(program, "addVector", &ocl_error);
	if (ocl_error != CL_SUCCESS) quit("Error creating kernel");

	//Set kernel arguments
	for (int i = 0;i < 3;i++)
	{
		if (clSetKernelArg(kernel, i, sizeof(cl_mem), &device_vector[i]) != CL_SUCCESS) quit("Error setting kernel argument");
	}
	int vector_size = TEST_VECTOR_SIZE;
	if (clSetKernelArg(kernel, 3, sizeof(int), &vector_size) != CL_SUCCESS) quit("Error setting kernel argument");

	//Execute Kernel
	size_t local_size = WORK_GROUP_SIZE;
	size_t global_size = WORK_GROUP_SIZE;
	if (clEnqueueNDRangeKernel(command_queue, kernel, 1, NULL, &global_size, &local_size, 0, NULL, NULL) != CL_SUCCESS) quit("Error executing kernel");

	//Fetch result from device, the command is blocking so this thread will only continue after the data was fetched from the device
	if (clEnqueueReadBuffer(command_queue, device_vector[2], CL_TRUE, 0, TEST_VECTOR_SIZE * sizeof(int), host_vector[2], 0, NULL, NULL) != CL_SUCCESS) quit("Error reading result from device");

	//Display result
	printf("Result: ");
	for (int i = 0;i < TEST_VECTOR_SIZE;i++)
	{
		if (i == 64) break;
		printf("%d ", host_vector[2][i]);
	}
	printf("\n");

	//Clean Up
	clReleaseKernel(kernel);
	clReleaseProgram(program);
	for (int i = 0;i < 3;i++)
	{
		delete[] host_vector[i];
		clReleaseMemObject(device_vector[i]);
	}
	clReleaseCommandQueue(command_queue);
	clReleaseContext(context);

	//Exit
	return 0;
}

